/** \file element.hh
 * Element class, categories and properties. */
/*
 * This file is part of libgelemental, a periodic table library with detailed
 * information on elements.
 *
 * Copyright (C) 2006-2007 Kevin Daughtridge <kevin@kdau.com>
 * Copyright (C) 2003 Jonas Frantz <jonas.frantz@welho.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef LIBGELEMENTAL__ELEMENT_HH
#define LIBGELEMENTAL__ELEMENT_HH

#include <libgelemental/value.hh>
#include <libgelemental/value-types.hh>

#include <list>
#include <ostream>
#include <vector>

/******************************************************************************/
namespace gElemental {
/******************************************************************************/

#ifndef DOXYGEN_SHOULD_SKIP_THIS
struct ElementData; /* defined in data.hh */
#endif

/******************************************************************************/
/** A property of the chemical elements. */
struct PropertyBase {
/******************************************************************************/

	/** Constructs a new property. */
	PropertyBase(const Message& name, const Message& format,
		const Message& description, const Message* sources[]);

	/** The name of the property. */
	Message m_name;
	
	/** Returns the translated name of the property. */
	ustring get_name() const;
	
	/** A compose-style format for output of values. */
	Message m_format;

	/** An explanatory description of the property. */
	Message m_description;
	
	/** Citations for the sources of values for the property. */
	std::list<const Message*> m_sources;

    /** Possibly adds a value for this property to an EntriesView.
	 * An entry will only be made if the value is defined or if its \p m_always
	 * is \c true.
	 * \param view The EntriesView to be populated.
	 * \param value The value to be added. */
	void make_entry(EntriesView& view, const value_base& value) const;

    /** Adds a string representing a value for this property to an EntriesView.
	 * \param view The EntriesView to be populated.
	 * \param value The value string to be added. */
	void make_entry(EntriesView& view, const ustring& value) const;

}; /* struct PropertyBase */

/******************************************************************************/
/** A typed property of the chemical elements. */
template<class T> struct Property : public PropertyBase {
/******************************************************************************/

	Property(const Message& name, const Message& format, const Message& desc,
		const Message* sources[])
	:	PropertyBase(name, format, desc, sources)
	{}

}; /* struct Property */

/******************************************************************************/
/** A category of properties of the chemical elements. */
struct Category {
/******************************************************************************/

	/** Constructs a new category. */
	Category(const Message& name, PropertyBase* properties[]);

	/** The name of the category. */
	Message m_name;
	
	/** Returns the translated name of the category. */
	ustring get_name() const;
	
	/** The properties in the category. */
	std::list<PropertyBase*> m_properties;
	
	/** Adds the category name to an EntriesView.
	* \param view The EntriesView to be populated. */
	void make_header(EntriesView& view) const;

}; /* struct Category */

/******************************************************************************/
/** A chemical element.
 * The element properties are named in the Properties sections of the namespace.
 * For details, see the Properties Information dialog. */
class Element {
/******************************************************************************/

public:
	
	/** Returns the value_base of the value for the given Property.
	 * \param property The Element Property to be returned. */
	const value_base& get_property_base(const PropertyBase& property) const;

	/** Returns the value for the given Property.
	 * \param property The Element Property to be returned. */
	template<class T>
	const T& get_property(const Property<T>& property) const
	{	return *reinterpret_cast<const T*>(&get_property_base(property)); }

	/** Generates entries for all properties.
	 * \param view An EntriesView to be populated. */
	void make_entries(EntriesView& view) const;

	/** Generates entries for the properties in the given Category.
	 * \param view An EntriesView to be populated.
	 * \param category The Category whose properties will be provided.
	 * \param all For \p C_GENERAL, controls whether the EntriesView will also
	 * receive the element's primary name, symbol, and atomic number. */
	void make_entries(EntriesView& view, const Category& category,
		bool all = false) const;

	/** Prints a full table of entries on the element to an output stream.
	 * \param os The output stream which will receive the entries. */
	void print_entries(std::ostream& os) const;

	/** Returns the phase of matter assumed by the element at standard pressure
	 * and a given temperature.
	 * \param tempK The temperature, in Kelvin, at which to determine phase. By
	 * default, the IUPAC standard temperature.
	 * \return The phase of matter determined. */
	Phase get_phase(double tempK = STANDARD_TEMPERATURE) const;

	/** The chemical symbol for this element. */
	ustring m_symbol;
	
	/** The atomic number of this element. */
	unsigned int m_number;

#ifndef DOXYGEN_SHOULD_SKIP_THIS

private:

	friend void initialize();
	Element(const ElementData& data);

	const ElementData& m_data;
	vustring m_official_name;
	Phase m_standard_phase;
	vfloat m_lattice_volume;
	
	vfloat get_lattice_volume() const;

#endif

}; /* class Element */

/******************************************************************************/
/** The categories of available properties of the chemical elements. */
/******************************************************************************/

extern std::list<Category*> CATEGORIES;

/******************************************************************************/
/** \name General Properties
 * @{ */
/******************************************************************************/

extern Category C_GENERAL;
/** Translators: This should be translated to the name most commonly used in
 * scientific contexts in your language. If this is different from the
 * official name, the official name will be presented untranslated in a
 * second entry. */
extern Property<Message> P_NAME;
extern Property<vustring> P_OFFICIAL_NAME;
/** Translators: This name is not available for translation. It will not be
 * presented if it is equal to the translated primary name. */
extern Property<vustring> P_ALTERNATE_NAME;
extern Property<ustring> P_SYMBOL;
extern Property<unsigned int> P_NUMBER;	
extern Property<Series> P_SERIES;
extern Property<vint> P_GROUP;
extern Property<vint> P_PERIOD;
extern Property<Block> P_BLOCK;
/** @} */

/******************************************************************************/
/** \name Historical Properties
 * @{ */
/******************************************************************************/

extern Category C_HISTORICAL;
extern Property<Event> P_DISCOVERY;
extern Property<Message> P_DISCOVERED_BY;
/** Translators: See po/README for translation issues. */
extern Property<Message> P_ETYMOLOGY;
/** @} */

/******************************************************************************/
/** \name Physical Properties
 * @{ */
/******************************************************************************/

extern Category C_PHYSICAL;
extern Property<Phase> P_PHASE;
extern Property<vfloat> P_DENSITY_SOLID;
extern Property<vfloat> P_DENSITY_LIQUID;
extern Property<vfloat> P_DENSITY_GAS;
extern Property<Message> P_APPEARANCE;
/** @} */

/******************************************************************************/
/** \name Thermal Properties
 * @{ */
/******************************************************************************/

extern Category C_THERMAL;
extern Property<vfloat> P_MELTING_POINT;
extern Property<vfloat> P_BOILING_POINT;
extern Property<vfloat> P_FUSION_HEAT;
extern Property<vfloat> P_VAPORIZATION_HEAT;
extern Property<vfloat> P_SPECIFIC_HEAT;
extern Property<vfloat> P_THERMAL_CONDUCTIVITY;
extern Property<vfloat> P_DEBYE_TEMPERATURE;
/** @} */

/******************************************************************************/
/** \name Atomic Properties
 * @{ */
/******************************************************************************/

extern Category C_ATOMIC;
extern Property<vfloat> P_ATOMIC_MASS;
extern Property<vfloat> P_ATOMIC_VOLUME;
extern Property<vfloat> P_ATOMIC_RADIUS;
extern Property<vfloat> P_COVALENT_RADIUS;
extern Property<vfloat> P_VAN_DER_WAALS_RADIUS;
extern Property<vustring> P_IONIC_RADII;
/** @} */

/******************************************************************************/
/** \name Crystallographic Properties
 * @{ */
/******************************************************************************/

extern Category C_CRYSTALLOGRAPHIC;
extern Property<Lattice> P_LATTICE_TYPE;
extern Property<vint> P_SPACE_GROUP;
extern Property<vfloatlist> P_LATTICE_EDGES;
extern Property<vfloatlist> P_LATTICE_ANGLES;
extern Property<vfloat> P_LATTICE_VOLUME;
/** @} */

/******************************************************************************/
/** \name Electronic Properties
 * @{ */
/******************************************************************************/

extern Category C_ELECTRONIC;
/** Superscripts should be denoted by the tag <tt>\<sup\>\</sup\></tt>. */
extern Property<vustring> P_CONFIGURATION;
extern Property<vintlist> P_OXIDATION_STATES;
extern Property<vfloat> P_ELECTRONEGATIVITY;
extern Property<vfloat> P_ELECTRON_AFFINITY;
extern Property<vfloat> P_FIRST_ENERGY;
/** @} */

/******************************************************************************/
/** \name Miscellaneous Properties
 * @{ */
/******************************************************************************/

extern Category C_MISCELLANEOUS;
extern Property<ColorValue> P_COLOR;
extern Property<Message> P_NOTES;
/** @} */

} /* namespace gElemental */

#endif /* LIBGELEMENTAL__ELEMENT_HH */
