/** \file value-types.hh
 * Instantiations of value<> and higher-level value types. */
/*
 * This file is part of libgelemental, a periodic table library with detailed
 * information on elements.
 *
 * Copyright (C) 2006-2007 Kevin Daughtridge <kevin@kdau.com>
 * Copyright (C) 2003 Jonas Frantz <jonas.frantz@welho.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef LIBGELEMENTAL__VALUE_TYPES_HH
#define LIBGELEMENTAL__VALUE_TYPES_HH

#include <libgelemental/value.hh>

/******************************************************************************/
namespace gElemental {
/******************************************************************************/

/** A qualified floating-point value. */
typedef value<double> vfloat;

/** A qualified integer value. */
typedef value<int> vint;

/** A qualified UTF-8 string value that does not require translation. */
typedef value<ustring> vustring;

/** A qualified list of floating-point values. */
typedef value_list<double> vfloatlist;

/** A qualified list of integer values. */
typedef value_list<int> vintlist;

/******************************************************************************/
/** A qualified UTF-8 string value that requires translation. */
class Message : public value_base {
/******************************************************************************/

public:

	/** Constructs a new defined message.
	 * \param src The untranslated message.
	 * \param qualifier The qualification of the message. */
	Message(const char* src, Qualifier qualifier = Q_NEUTRAL);

	/** Constructs a new undefined or empty message.
	 * \param qualifier The qualification of the message. */
	Message(Qualifier qualifier = Q_UNK);

	virtual ~Message();

	/** Compares two messages.
	 * \param rhs The message for comparison.
	 * \return \c -1 if this message is undefined or sorts before \p rhs; \c 0
	 * if the messages sort equally; or \c 1 if \p rhs is undefined or this
	 * message sorts after \p rhs. */
	virtual int compare(const value_base& rhs) const;

	virtual ustring get_string(const ustring& format = ustring()) const;

	/** The untranslated message, if defined. */
	ustring val;

protected:

	/** Returns a localized text representation of a defined message.
	 * \return A UTF-8 string containing the translated message. */
	virtual ustring do_get_string(const ustring& format) const;

}; /* class Message */

/******************************************************************************/
/** A qualified value representing the time and location of an event. */
class Event : public value_base {
/******************************************************************************/

public:

	/** Constructs a new defined event.
	 * \param when The year in which the event occurred.
	 * \param where The place in which the event occurred.
	 * \param qualifier The qualification of the event. */
	Event(int when, const ustring& where, Qualifier qualifier = Q_NEUTRAL);

	/** Constructs a new undefined or empty event.
	 * \param qualifier The qualification of the event. */
	Event(Qualifier qualifier = Q_UNK);

	virtual ~Event();

	/** Compares two events.
	 * \param rhs The event for comparison.
	 * \return \c -1 if this event is undefined or occurred before \p rhs; \c 0
	 * if the events occurred in the same year; or \c 1 if \p rhs is undefined
	 * or this event occurred after \p rhs. */
	virtual int compare(const value_base& rhs) const;

	/** Returns a localized text representation of the event.
	 * \param format An optional compose-style format string with which to
	 * output the value. The year is "%1" and the place is "%2". By default,
	 * the format is the localized version of \c "%1 (%2)".
	 * \return A UTF-8 string containing the localized representation. */
	virtual ustring get_string(const ustring& format = ustring()) const;
	
	/** The year in which the event occurred, if defined. */
	int m_when;
	
	/** The place in which the event occurred, if defined. */
	ustring m_where;

protected:

	/** Returns a localized text representation of a defined event.
	 * \return A UTF-8 string containing the localized representation. */
	virtual ustring do_get_string(const ustring& format) const;

	/** Compares the qualifications of two events. */
	virtual int compare_base(const value_base& rhs) const;

}; /* class Event */

/******************************************************************************/
/** A qualified value representing a chemical series. */
class Series : public color_value {
/******************************************************************************/

public:

	/** The possible chemical series values. */
	enum Value {
		NONMETAL,
		NOBLE_GAS,
		ALKALI_METAL,
		ALKALINE_EARTH_METAL,
		SEMIMETAL,
		HALOGEN,
		POST_TRANSITION_METAL,
		TRANSITION_METAL,
		LANTHANIDE,
		ACTINIDE
	};

	/** Constructs a new defined series.
	 * \param src The initial definition.
	 * \param qualifier The qualification of the initial value. */
	Series(const Value& src, Qualifier qualifier = Q_NEUTRAL);

	/** Constructs a new undefined or default series.
	 * \param qualifier The qualification of the initial value. */
	Series(Qualifier qualifier = Q_UNK);
	
	virtual ~Series();

	/** Compares two series.
	 * \param rhs The series for comparison.
	 * \return \c -1 if this series is undefined or sorts before \p rhs; \c 0 if
	 * the series are equal; or \c 1 if \p rhs is undefined or this series sorts
	 * after \p rhs. */
	virtual int compare(const value_base& rhs) const;

	virtual color get_color() const;
	
	/** The value, if defined. */
	Value val;

protected:

	/** Returns a localized text description of the defined value.
	 * \return A UTF-8 string naming the series. */
	virtual ustring do_get_string(const ustring& format) const;

}; /* class Series */

/******************************************************************************/
/** A qualified value representing a period table block. */
class Block : public color_value {
/******************************************************************************/

public:

	/** The possible lattice type values. */
	enum Value {
		S, /**< s-block (sharp) */
		P, /**< p-block (principal) */
		D, /**< d-block (diffuse) */
		F, /**< f-block (fundamental) */
	};

	/** Constructs a new defined block.
	 * \param src The initial definition.
	 * \param qualifier The qualification of the initial value. */
	Block(const Value& src, Qualifier qualifier = Q_NEUTRAL);

	/** Constructs a new undefined or default block.
	 * \param qualifier The qualification of the initial value. */
	Block(Qualifier qualifier = Q_UNK);
	
	virtual ~Block();

	/** Compares two blocks.
	 * \param rhs The blocks for comparison.
	 * \return \c -1 if this block is undefined or sorts before \p rhs; \c 0 if
	 * the blocks are equal; or \c 1 if \p rhs is undefined or this block sorts
	 * after \p rhs. */
	virtual int compare(const value_base& rhs) const;

	virtual color get_color() const;

	/** The value, if defined. */
	Value val;

protected:

	/** Returns a localized text description of the defined value.
	 * \return A UTF-8 string naming the lattice type. */
	virtual ustring do_get_string(const ustring& format) const;

}; /* class Block */

/******************************************************************************/
/** A qualified value representing a phase of matter. */
class Phase : public color_value {
/******************************************************************************/

public:

	/** The possible phase values. */
	enum Value {
		SOLID,
		LIQUID,
		GAS
	};

	/** Constructs a new defined phase.
	 * \param src The initial definition.
	 * \param qualifier The qualification of the initial value. */
	Phase(const Value& src, Qualifier qualifier = Q_NEUTRAL);

	/** Constructs a new undefined or default phase.
	 * \param qualifier The qualification of the initial value. */
	Phase(Qualifier qualifier = Q_UNK);
	
	virtual ~Phase();

	/** Compares two phases.
	 * \param rhs The phases for comparison.
	 * \return \c -1 if this phase is undefined or sorts before \p rhs; \c 0 if
	 * the phases are equal; or \c 1 if \p rhs is undefined or this phase sorts
	 * after \p rhs. */
	virtual int compare(const value_base& rhs) const;

	virtual color get_color() const;

	/** The value, if defined. */
	Value val;

protected:

	/** Returns a localized text description of the defined value.
	 * \return A UTF-8 string naming the phase. */
	virtual ustring do_get_string(const ustring& format) const;

}; /* class Phase */

/******************************************************************************/
/** A qualified value representing a Bravais lattice type. */
class Lattice : public color_value {
/******************************************************************************/

public:

	/** The possible lattice type values. */
	enum Value {
		TRI, /**< triclinic */
		MONO, /**< monoclinic */
		ORTH, /**< orthorhombic */
		TET, /**< tetragonal */
		RHO, /**< rhombohedral */
		HEX, /**< hexagonal */
		SC, /**< simple cubic */
		BCC, /**< body-centered cubic */
		FCC /**< face-centered cubic */
	};

	/** Constructs a new defined lattice type.
	 * \param src The initial definition.
	 * \param qualifier The qualification of the initial value. */
	Lattice(const Value& src, Qualifier qualifier = Q_NEUTRAL);

	/** Constructs a new undefined or default lattice type.
	 * \param qualifier The qualification of the initial value. */
	Lattice(Qualifier qualifier = Q_UNK);
	
	virtual ~Lattice();

	/** Compares two lattices.
	 * \param rhs The lattices for comparison.
	 * \return \c -1 if this lattice is undefined or sorts before \p rhs; \c 0
	 * if the lattices are equal; or \c 1 if \p rhs is undefined or this lattice
	 * sorts after \p rhs. */
	virtual int compare(const value_base& rhs) const;

	virtual color get_color() const;

	/** The value, if defined. */
	Value val;

protected:

	/** Returns a localized text description of the defined value.
	 * \return A UTF-8 string naming the lattice type. */
	virtual ustring do_get_string(const ustring& format) const;

}; /* class Lattice */

/******************************************************************************/
/** A qualified value representing a display color. */
class ColorValue : public color_value {
/******************************************************************************/

public:

	/** Constructs a new defined color.
	 * \param src The initial definition.
	 * \param qualifier The qualification of the initial value. */
	ColorValue(const color& src, Qualifier qualifier = Q_NEUTRAL);

	/** Constructs a new undefined or default color.
	 * \param qualifier The qualification of the initial value. */
	ColorValue(Qualifier qualifier = Q_UNK);
	
	virtual ~ColorValue();

	/** The display color, if defined. */
	color val;

	/** Compares two colors.
	 * \param rhs The colors for comparison.
	 * \return \c -1 if this color is undefined or is darker than \p rhs; \c 0
	 * if the colors are equal in value; or \c 1 if \p rhs is undefined or this
	 * color is brighter than \p rhs. */
	virtual int compare(const value_base& rhs) const;

	/** Returns a localized text representation of the color.
	 * \param format An optional compose-style format string with which to
	 * output the value. The red value is "%1", the green "%2", and the blue
	 * "%3", each suitable for percentage form (multiplied by 100).
	 * \return A UTF-8 string containing the localized representation. */
	virtual ustring get_string(const ustring& format = ustring()) const;

	virtual color get_color() const;

protected:

	/** Returns a localized text description of the defined value.
	 * \return A UTF-8 string naming the phase. */
	virtual ustring do_get_string(const ustring& format) const;

}; /* class ColorValue */

} /* namespace gElemental */

#endif /* LIBGELEMENTAL__VALUE_TYPES_HH */
