/** \file value.hh
 * Basic definitions and structures, and base value classes. */
/*
 * This file is part of libgelemental, a periodic table library with detailed
 * information on elements.
 *
 * Copyright (C) 2006-2007 Kevin Daughtridge <kevin@kdau.com>
 * Copyright (C) 2003 Jonas Frantz <jonas.frantz@welho.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef LIBGELEMENTAL__VALUE_HH
#define LIBGELEMENTAL__VALUE_HH

#include <libgelemental/compose/ucompose.hh>
#include <libgelemental/compose/extras.hh>

#include <iomanip>
#include <vector>
#include <glibmm/refptr.h>
#include <glibmm/ustring.h>

/******************************************************************************/
/** A periodic table library with detailed information on elements. */
namespace gElemental {
/******************************************************************************/

using Glib::RefPtr;
using Glib::ustring;

/** The standard temperature, in Kelvin, as defined by IUPAC. */
extern const double STANDARD_TEMPERATURE;

/** A qualification of the meaning of a value. */
enum Qualifier {
	Q_NEUTRAL, /**< Indicates a presumably valid value. */
	Q_UNK, /**< Indicates that a value is unknown or unavailable. */
	Q_NA, /**< Indicates that a value is not applicable for an element. */
	Q_EST, /**< Indicates that a value is estimated or calculated. */
	Q_CA, /**< Indicates that a value is approximate. */
	Q_ISO /**< Indicates a value for the most stable isotope of an element. */
};

/******************************************************************************/
/** An interface for the display of value entries. */
class EntriesView {
/******************************************************************************/

public:

	virtual ~EntriesView();

	/** Displays a category header.
	 * \param category The name of the category. */
	virtual void header(const ustring& category) = 0;

	/** Displays a single value entry.
	 * \param name The name of the entry.
	 * \param value A localized string representation of the value of the entry.
	 * \param tip Qualifying information, if any, to be displayed as a tip. */
	virtual void entry(const ustring& name, const ustring& value,
		const ustring& tip = ustring()) = 0;

}; /* class EntriesView */

/******************************************************************************/
/** A base class for qualified values. */
class value_base {
/******************************************************************************/

public:

	virtual ~value_base();

	/** The qualification of the value's meaning. */
	Qualifier m_qualifier;
	
	/** Returns whether the value is defined.
	 * \return \c false if the value is unknown or inapplicable; \c true
	 * otherwise */
	bool has_value() const;

	/** Returns a localized text representation of the value.
	 * \param format An optional compose-style format string with which to
	 * output the value. If not otherwise specified, the value is argument
	 * reference \c %%1.
	 * \return A UTF-8 string containing the localized representation. */
	virtual ustring get_string(const ustring& format = ustring()) const;

	/** Returns a localized message to be displayed as a tooltip or
	 * parenthetically to the value.
	 * \return A UTF-8 string containing the tip text, if any. */
	ustring get_tip() const;

	/** Possibly adds the result of get_string() to an EntriesView.
	 * An entry will only be made if the value is defined or if \p m_always is
	 * \c true.
	 * \param view The EntriesView to be populated.
	 * \param name The name of the value to be passed.
	 * \param format An optional compose-style format for get_string(). */
	void make_entry(EntriesView& view, const ustring& name,
		const ustring& format = ustring()) const;

	/** Compares two qualified values.
	 * Derived classes may override this function, first calling compare_base,
	 * and use a dynamic_cast to check rhs.
	 * \param rhs The qualified value for comparison.
	 * \return \c -1 if this value is undefined or is less than \p rhs; \c 0 if
	 * the values are equal; or \c 1 if \p rhs is undefined or this value is
	 * greater than \p rhs. */
	virtual int compare(const value_base& rhs) const;

protected:

	/** Constructs a new value_base.
	 * \param qualifier The qualification of the initial value. */
	value_base(Qualifier qualifier);

	/** Returns a localized text representation of a defined value.
	 * Derived classes must implement this function to produce a text
	 * representation based on the format. It will be called by get_string(),
	 * which will then decorate its output as appropriate for the current
	 * qualifier. If has_value() is \c false, the behavior is undefined.
	 * \param format A compose-style format string. Derived classes may
	 * define the number of argument references expected in the string; the
	 * default is one. If \p format is empty, a minimal default should be used.
	 * \return A UTF-8 string containing the localized representation. */
	virtual ustring do_get_string(const ustring& format) const = 0;

	/** Compares the qualifications of two values.
	 * This base function handles definedness only. Derived classes may
	 * override this function to handle qualifiers more specifically.
	 * \param rhs The qualified value for comparison.
	 * \return \c -1 if this value sorts before \p rhs; \c 0 if the values
	 * sort equally; \c 1 if this value sorts after \p rhs; or \p YIELD_COMPARE
	 * if the result should be determined based on values. */
	virtual int compare_base(const value_base& rhs) const;
	static const int YIELD_COMPARE;

	/** Whether an undefined value should make an entry.
	 * May be set by derived classes.
	 * If set to \c false (default), make_entry() will only populate an
	 * EntriesView if has_value() returns \c true. If set to \c true, an
	 * EntriesView will always be populated. */
	bool m_always;

}; /* class value_base */

/******************************************************************************/
/** A color description.
 * Its components are suitable as parameters to Gdk::Color::set_rgb_p(). */
struct color {
/******************************************************************************/

	/** Constructs a new color with red, green, and blue percentages. */
	color(double red, double green, double blue);

	/** Constructs a new color with red, green, and blue levels.
	 * Each component ranges from 0 (0.0) to 65535 (1.0). */
	color(unsigned int red, unsigned int green, unsigned int blue);

	/** The red component of the color in the RGB space. */
	double red;

	/** The green component of the color in the RGB space. */
	double green;

	/** The blue component of the color in the RGB space. */
	double blue;

	/** Returns the value component of the color in the HSV space. */
	double get_hsv_value() const;

	/** Returns a foreground color complimentary to the color as background. */
	color get_compliment() const;

}; /* struct color */

/******************************************************************************/
/** A base class for qualified values which have representative colors. */
class color_value : public value_base {
/******************************************************************************/

public:

	virtual ~color_value();

	/** Returns a color representing the value.
	 * Derived classes must implement this function, returning an allocated
	 * GDK color that corresponds to the defined value. A special color should
	 * be returned if has_value() is \c false.
	 * \return A GDK color value. */
	virtual color get_color() const = 0;

protected:

	/** Constructs a new color_value.
	 * \param qualifier The qualification of the initial value. */
	color_value(Qualifier qualifier);

}; /* class color_value */

/******************************************************************************/
/** A qualified value of a generic type.
 * This template should only be instantiated for classes with a default
 * constructor, a copy constructor, a destructor, and less-than, greater-than,
 * and stream insertion operators. Compiler-generated functions are usable. */
template<class T>
class value : public value_base {
/******************************************************************************/

public:

	/** Constructs a new defined value.
	 * \param src The initial definition.
	 * \param qualifier The qualification of the initial value. */
	template<class S>
	value(const S& src, Qualifier qualifier = Q_NEUTRAL)
	:	value_base(qualifier), val(src)
	{}

	/** Constructs a new value based on another value.
	 * \param src The value to be copied. */
	template<class S>
	value(const value<S>& src)
	:	value_base(src.m_qualifier), val(src.val)
	{}

	/** Constructs a new undefined or default value.
	 * \param qualifier The qualification of the initial value. */
	value(Qualifier qualifier = Q_UNK)
	:	value_base(qualifier), val()
	{}
	
	virtual
	~value()
	{}

	virtual int
	compare(const value_base& rhs) const
	{
		int base = compare_base(rhs);
		if (base != YIELD_COMPARE) return base;
		
		const value<T> *vrhs = dynamic_cast<const value<T>*>(&rhs);
		if (vrhs != NULL) {
			if (val < vrhs->val)
				return -1;
			else if (val > vrhs->val)
				return 1;
			else
				return 0;
		}

		return 0;
	}

	/** The value, if defined. */
	T val;

protected:

	/** Returns a localized text representation of the defined value.
	 * \return The UTF-8 string result of a stream insertion of the value. */
	virtual ustring
	do_get_string(const ustring& format) const
	{
		if (format.empty())
			return compose::ucompose1(val, std::numeric_limits<T>::digits10);
		else
			return compose::UComposition(format)
				.precision(std::numeric_limits<T>::digits10)
				.arg(val).str();
	}

}; /* class value */

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/** Returns a series of characters separating members of a list in output.
 * For internal use. */
ustring get_list_separator();
#endif

/******************************************************************************/
/** A qualified list of values of a generic type.
 * This template should only be instantiated for classes with a default
 * constructor, a copy constructor, a destructor, and less-than, greater-than,
 * and stream insertion operators. Compiler-generated functions are usable. */
template<class T>
class value_list : public value_base {
/******************************************************************************/

public:

	/** Constructs a new defined value list.
	 * \param src An array of the initial values.
	 * \param count The length of the \p src array.
	 * \param qualifier The qualification of the initial values. */
	template<class S>
	value_list(const S src[], int count, Qualifier qualifier = Q_NEUTRAL)
	:	value_base(qualifier), vals()
	{
		for (int i = 0; i < count; ++i)
			vals.push_back(T(src[i]));
	}

	/** Constructs a new value list based on another value list.
	 * \param src The value list to be copied. */
	template<class S>
	value_list(const value_list<S>& src)
	:	value_base(src.m_qualifier), vals()
	{
		for (typename std::vector<S>::const_iterator i = src.vals.begin();
			i != src.vals.end(); ++i)
			vals.push_back(T(*i));
	}

	/** Constructs a new undefined or empty value list.
	 * \param qualifier The qualification of the initial values. */
	value_list(Qualifier qualifier = Q_UNK)
	:	value_base(qualifier), vals()
	{}
	
	virtual
	~value_list()
	{}

	virtual int
	compare(const value_base& rhs) const
	{
		int base = compare_base(rhs);
		if (base != YIELD_COMPARE) return base;
		
		const value_list<T> *vrhs = dynamic_cast<const value_list<T>*>(&rhs);
		if (vrhs != NULL) {
			if (vals < vrhs->vals)
				return -1;
			else if (vals > vrhs->vals)
				return 1;
			else
				return 0;
		}

		return 0;
	}

	/** The values, if defined. */
	mutable std::vector<T> vals;

protected:

	/** Returns a localized text representation of the defined value list.
	 * \return The UTF-8 string result of a concatenation of the values. */
	virtual ustring
	do_get_string(const ustring& format) const
	{
		ustring result;
		for (typename std::vector<T>::const_iterator i = vals.begin();
			i != vals.end(); ++i) {
			if (i != vals.begin()) result += get_list_separator();
			if (format.empty())
				result += compose::ucompose1(*i,
					std::numeric_limits<T>::digits10);
			else
				result += compose::UComposition(format)
					.precision(std::numeric_limits<T>::digits10)
					.arg(*i).str();
		}
		return result;
	}

}; /* class value_list */

} /* namespace gElemental */

#endif /* LIBGELEMENTAL__VALUE_HH */
